with Ada.Text_IO, Strings, Ints, Flts, Globals, Params, Reps.Balls;
use Ada.Text_IO, Strings, Ints, Flts, Globals, Params;
with Vectors.Ops, Matrices, ScalVectors, Polynoms, Taylors1, Rattleback;
with ScalVectors.Eigenv, ScalVectors.Eigenv.RNum;

pragma Elaborate_All (Ada.Text_IO,Strings,Ints,Flts,Globals,Params,Reps);
pragma Elaborate_All (Vectors,Polynoms,Taylors1);

procedure Basin is --- estimate of the basin of attraction

  NDigits:      constant Positive := 1024;
  Dummy:        constant Positive := Reps.Set_Precision(NDigits);
  Rad:          constant Radius   := One;   -- for Taylors1
  PDeg:         constant Positive := 9;     -- Taylor degree
  ABFlag:       constant Flag     := 0;     -- Alpha,Beta undefined
  FRP:          constant FRParam  := Param13;

  type FMatrix4 is array(1..4,1..4) of Flt;

  procedure MakeMat(O3: in Flt; M: in out FMatrix4) is
    use Reps;
    subtype NumScalar is Rep;
    package RB is new Rattleback(FRP => FRP, ABFlag => ABFlag, Scalar => NumScalar);
    package SM is new Matrices (Scalar => NumScalar);
    package SV is new Vectors (Component => NumScalar);
    package SSV is new ScalVectors (Scalar => NumScalar, SM => SM, SV => SV);
    package SSV_EV is new SSV.Eigenv;
    package SSV_EV_RNum is new SSV_EV.RNum;
    subtype SVector4 is SV.Vector(1..4);
    subtype SMatrix4 is SM.Matrix(1..4,1..4);
    Omega3: NumScalar;
    Wr,Wi: SVector4;
    C: SVector4 renames Wi;
    Vr,Vi: SMatrix4;
    DX: RB.SMatrix4;
    procedure SetColumn(J: in Integer; Fac: in Radius; V: in SVector4; A: in out FMatrix4) is
    begin
      for I in 1..4 loop A(I,J) := Fac*Approx(V(I)); end loop; -- Fac is balancing A
    end SetColumn;
    use RB,SSV,SSV_EV_RNum;
  begin
    Assign(O3,Omega3);
    DVecField_Stat(Omega3,DX);
    Eigen(SMatrix4(DX),Wr,Wi,Vr,Vi);
    Show2("EV: approximate real parts",Approx(Wr(1)),Approx(Wr(3)));
    GetColumn(1,Vr,C); SetColumn(1,One,C,M);
    GetColumn(1,Vi,C); SetColumn(2,One,C,M);
    GetColumn(3,Vr,C); SetColumn(3,Three,C,M);
    GetColumn(3,Vi,C); SetColumn(4,Three,C,M);
  end MakeMat;

  use Reps.Balls;
  subtype Scalar is Ball;
  package SM is new Matrices (Scalar => Scalar);
  package SV is new Vectors (Component => Scalar);
  package SVO is new SV.Ops (Scalar => Scalar, SM => SM);
  package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);
  package SP is new Polynoms (Scalar => Scalar, SV => SV);
  package Tays is new Taylors1 (PDeg => PDeg, Rad => Rad, Scalar => Scalar, SV =>  SV, SP => SP);
  use SM,SV,SVO,Tays;
  package SRB is new Rattleback(FRP => FRP, ABFlag => ABFlag, Scalar => Scalar);
  package TRB is new Rattleback(FRP => FRP, ABFlag => ABFlag, Scalar => Taylor1);

  subtype SVector4 is SV.Vector(1..4);
  subtype SMatrix4 is SM.Matrix(1..4,1..4);

  procedure Inner_Product(InvMat: in SMatrix4; U,V: in SVector4; S: in out Scalar) is
    Ut,Vt: SVector4;
  begin
    Prod(InvMat,U,Ut);
    Prod(InvMat,V,Vt);
    SSV.Inner_Prod(Ut,Vt,S);
  end Inner_Product;

  procedure Ball_Enclosure(Mat,InvMat: in Matrix; X: in SVector4; R: out Radius; V: in out SVector4) is
    S,Tmp: Scalar;
    U: SVector4;
  begin
    Inner_Product(InvMat,X,X,Tmp);
    Sqrt(Tmp,S);
    R := MaxNorm(S);
    for I in 1..4 loop BallAt0(R,U(I)); end loop;
    Prod(Mat,U,V);
  end Ball_Enclosure;

  procedure Show_NonLinear(X: in SVector4; O3,En: in Scalar; Mat,InvMat: in SMatrix4) is
    use TRB;
    R,SupNLin: Radius;
    SOne,M3,IProd: Scalar;
    DotGamma,DotM: SVector3;
    BX,FBX: SVector4;
    S,G1,G2: Taylor1;
    PT: Point;
  begin
    Ball_Enclosure(Mat,InvMat,X,R,BX);
    Assign(One,SOne);
    SetZero(S);
    AddCoeff(1,SOne,S);
    Prod(BX(1),S,G1);
    Prod(BX(2),S,G2);
    MakeG3(G1,G2,PT.Gamma);
    Prod(BX(3),S,PT.M(1));
    Prod(BX(4),S,PT.M(2));
    Prod(FRP.I33,O3,M3);
    Assign(M3,PT.M(3));
    Assign(En,S);
    MakeM3(S,PT);            -- fix M3
    VecField(PT.Gamma,PT.M,DotGamma,DotM);
    for I in OneTwo loop
      SetZero(DotGamma(I).C(1));
      SetZero(DotM(I).C(1));
    end loop;
    Val(DotGamma(1),SOne,FBX(1));
    Val(DotGamma(2),SOne,FBX(2));
    Val(DotM(1),SOne,FBX(3));
    Val(DotM(2),SOne,FBX(4));
    Inner_Product(InvMat,X,FBX,IProd);
    SupNLin := MaxNorm(IProd);
    Show1("|<y,N(y)>/<y,y>| <",SupNLin*Sqr(One/R));
  end Show_NonLinear;

  function NormBound(A: SMatrix4) return Radius is
    use Flt_EF;
    R: Radius := Zero;
  begin
    for I in 1..4 loop
      for J in 1..4 loop R := R+Sqr(MaxNorm(A(I,J))); end loop;
    end loop;
    return Sqrt(R);
  end NormBound;

  use SRB;

  procedure Show_Linear(O3: in Scalar; Mat: in SMatrix4; InvMat: in out SMatrix4) is
    DX: SRB.SMatrix4;
    A,Tmp: SMatrix4;
    LambdaMin: Flt := Flt'Last;
    LambdaMax: Flt := -Flt'Last;
    SHalf: Scalar;
    Err: Radius;
  begin
    DVecField_Stat(O3,DX);
    Prod(SMatrix4(DX),Mat,Tmp);
    Prod(InvMat,Tmp,A);                    -- A = (1/Mat)*DF*Mat
    Transpose(A,Tmp);
    Add(Tmp,A);
    Assign(Half,SHalf);
    Mult(SHalf,A);
    for I in 1..4 loop
      LambdaMin := RMin(LambdaMin,Inf(A(I,I)));
      LambdaMax := RMax(LambdaMax,Sup(A(I,I)));
      SetZero(A(I,I));
    end loop;
    Err := NormBound(A);
    LambdaMax :=   Err+LambdaMax;
    LambdaMin := -(Err-LambdaMin);
    Show2("LambdaMin, LambdaMax",LambdaMin,LambdaMax);
  end Show_Linear;

  procedure Change(P: in Point; InvMat: in SMatrix4; X: in out SVector4) is
    MX: SVector4;
  begin
    Copy(P.Gamma(1),MX(1));
    Copy(P.Gamma(2),MX(2));
    Copy(P.M(1),MX(3));
    Copy(P.M(2),MX(4));
    Prod(InvMat,MX,X);
  end Change;

  procedure Check(FileName: in String) is
    use SRB.SMV;
    X: SVector4;
    E,B3,O3,T,Tmp: Scalar;
    M: FMatrix4;
    Mat,InvMat: SMatrix4;
    P: SRB.Point;
  begin
    Show0("Checking " & Filename & ":");
    Read(Input_Dir & Filename & ".init",P);
    Copy(P.Energy,E);
    Assign(FRP.BB3,Tmp);
    Sqrt(Tmp,B3);                          -- B3 = Sqrt(BB3)
    Prod(GForce,B3,T);                     -- m*g*B3
    Diff(E,T,Tmp);                         -- E - m*g*B3
    Prod(Two,Tmp,T);                       -- T = 2(E-M*g*B3)
    Quot(T,Inertia(3,3),Tmp);
    Sqrt(Tmp,O3);                          -- O3 = Sqrt(T/I33)
    Copy(O3,Tmp);
    Reduce_Precision(64,Tmp);
    Show1("Omega3 = ",Tmp);
    MakeMat(Approx(O3),M);
    for I in 1..4 loop
      for J in 1..4 loop Assign(M(I,J),Mat(I,J)); end loop;
    end loop;
    Copy(Mat,InvMat);
    SSV.Invert(InvMat);
    Show_Linear(O3,Mat,InvMat);
    Read(Input_Dir & Filename & ".final",P);
    Change(P,InvMat,X);
    Show_Nonlinear(X,O3,E,Mat,InvMat);
  end Check;

begin

  Check("P13-hetR");
  New_Line;
  Check("P13-hetRS");

end Basin;
