with Ada.Text_IO, Strings, Ints, Flts, Globals, Params, Reps.Balls;
use Ada.Text_IO, Strings, Ints, Flts, Globals, Params, Reps.Balls;
with Matrices, Vectors, Polynoms, Taylors1, MultiScalars, Small_Vectors, Rattleback;

pragma Elaborate_All (Ada.Text_IO,Strings,Ints,Flts,Globals,Params,Reps.Balls);
pragma Elaborate_All (Matrices,Vectors,Polynoms,Taylors1,Rattleback);

procedure Hurwitz is --- compute quantities needed for Routh-Hurwitz

  NDigits:      constant Positive := 1024;
  Dummy:        constant Positive := Reps.Set_Precision(NDigits);
  PDeg:         constant Positive := 10;    -- Taylor degree
  Rad:          constant Radius   := One;   -- for Taylors1
  ABFlag:       constant Flag     := 0;     -- Alpha,Beta undefined
  FRP:          constant FRParam  := Param13;

  subtype Scalar is Ball;
  package SM is new Matrices (Scalar => Scalar);
  package SV is new Vectors (Component => Scalar);
  package SP is new Polynoms (Scalar => Scalar, SV => SV);
  package STays is new Taylors1(PDeg => PDeg, Rad => Rad, Scalar => Scalar, SV =>  SV, SP => SP);
  subtype SPoly is STays.Poly1;
  subtype STay is STays.Taylor1; --- functions of Omega3 or M3
  use STays;

  procedure Make_Linear(S: in out STay) is
    --- S(x)=x
    C0,C1: Scalar;
  begin
    Assign(Zero,C0);
    Assign(One,C1);
    Affine(C0,C1,S);
  end Make_Linear;

  package TM is new Matrices (Scalar => STay);
  package TV is new Vectors (Component => STay);
  package TP is new Polynoms (Scalar => STay, SV => TV);
  package TTays is new Taylors1(PDeg => PDeg, Rad => Rad, Scalar => STay, SV => TV, SP => TP);
  subtype TPoly is TTays.Poly1;
  subtype TTay is TTays.Taylor1; --- functions of HatLambda or Lambda
  use TTays;

  procedure Make_Linear(T: in out TTay) is
    --- S(x)=x
    C0,C1: STay;
  begin
    Assign(Zero,C0);
    Assign(One,C1);
    Affine(C0,C1,T);
  end Make_Linear;

  procedure Make_HatCharPoly(T: in out TTay) is
    --- variables Omega3 and HatLambda
    package TTay_RB is new Rattleback(FRP => FRP, ABFlag => ABFlag, Scalar => TTay);
    use TTay_RB;
    ST: STay;
    Omega3,Lambda: TTay;
  begin
    Make_Linear(ST);
    Assign(ST,Omega3);
    Make_Linear(Lambda);
    HatCharPoly(Omega3,Lambda,T);
  end Make_HatCharPoly;

  procedure Minor1(C: in TPoly; D: in out STay) is
    Tmp: STay;
  begin
    Prod(C(3),C(2),D);
    Prod(C(1),C(4),Tmp);
    Sub(Tmp,D);
  end Minor1;

  procedure Minor2(C: in TPoly; D: in out STay) is
    package MS is new MultiScalars(Scalar => STay);
    package SMV is new Small_Vectors (Scalar => STay, MS => MS);
    M: MS.Matrix3;
  begin
    Copy(C(3),M(1,1));
    Copy(C(1),M(1,2));
    SetZero(  M(1,3));
    Copy(C(4),M(2,1));
    Copy(C(2),M(2,2));
    Copy(C(0),M(2,3));
    SetZero(  M(3,1));
    Copy(C(3),M(3,2));
    Copy(C(1),M(3,3));
    SMV.Det3(M,D);
  end Minor2;

  procedure Write_And_Show(Name: in String; S: in STay) is
    SR: STay;
  begin
    Write(Output_Dir & Name,S);
    if Verbosity>1 then
      Copy(S,SR);
      Reduce_Precision(64,SR);
      Show1(Name & " ",SR);
    end if;
  end Write_And_Show;

  procedure Show_Omega3(D: in STay) is
    Q,O3: Scalar;
  begin
    Quot(D.C(2),D.C(4),Q);
    Neg(Q);
    Sqrt(Q,O3);
    Reduce_Precision(64,O3);
    Show1("critical Omega3 = ",O3);
  end Show_Omega3;

  procedure Write_And_Show(T: in TTay) is
    C: TPoly renames T.C;
    D: STay;
  begin
    Write_And_Show("Coeff0",C(0));
    Write_And_Show("Coeff1",C(1));
    Write_And_Show("Coeff2",C(2));
    Write_And_Show("Coeff3",C(3));
    Write_And_Show("Coeff4",C(4));
    Minor1(C,D);
    Write_And_Show("Minor1",D);
    Minor2(C,D);
    Write_And_Show("Minor2",D);
    Show_Omega3(D);
  end Write_And_Show;

  T: TTay;

begin
  Verbosity := 2;

  Make_HatCharPoly(T);
  Write_And_Show(T);

end Hurwitz;
