with Globals;
use Globals;

pragma Elaborate_All(Globals);

package body Params is

  procedure Approx_Inertia(I1,I2,Del: in Flt; I11,I12,I22: out Flt) is
    use Flt_EF;
    C: constant Flt := Cos(Del);
    S: constant Flt := Sin(Del);
  begin
    I11 := C*C*I1+S*S*I2;
    I12 := C*S*(I1-I2);
    I22 := S*S*I1+C*C*I2;
  end Approx_Inertia;

  procedure Approx_Delta(I11,I12,I22: in Flt; I1,I2,Del: out Flt) is
  begin
    if I11>I22 then
      declare
        use Flt_EF;
        HalfTr: constant Flt := Half*(I11+I22);
        Det: constant Flt := I11*I22-I12*I12;
        Tmp: Flt := Sqrt(HalfTr*HalfTr-Det);
      begin
        I1 := HalfTr+Tmp;
        I2 := HalfTr-Tmp;
        Tmp := Two*I12/(I11-I22);
        Del := Half*Arctan(Tmp);
      end;
    elsif I11<I22 then
      Approx_Delta(I22,I12,I11,I2,I1,Del);
      Del := -Del;
    else
      I1 := I11;
      I2 := I22;
      Del := Zero;
    end if;
  end Approx_Delta;

  function Make_Approx_Param(I1,I2,I3,Del,BB1,BB2,BB3,Mass: Flt) return FRParam is
    I11,I12,I22: Flt;
  begin
    if BB1<BB2 then raise Not_Implemented; end if; --- undesirable
    if BB2<BB3 then raise Not_Implemented; end if; --- undesirable
    Approx_Inertia(I1,I2,Del,I11,I12,I22);
    return (I11,I12,I22,I3,BB1,BB2,BB3,Mass);
  end Make_Approx_Param;

  procedure Reduce_Precision(Binary_Digits: in Positive; P: in out FRParam) is
  begin
    Reduce_Precision(Binary_Digits,P.I11);
    Reduce_Precision(Binary_Digits,P.I12);
    Reduce_Precision(Binary_Digits,P.I22);
    Reduce_Precision(Binary_Digits,P.I33);
    Reduce_Precision(Binary_Digits,P.BB1);
    Reduce_Precision(Binary_Digits,P.BB2);
    Reduce_Precision(Binary_Digits,P.BB3);
    Reduce_Precision(Binary_Digits,P.Mass);
  end Reduce_Precision;

  function Approx_Homog_Param(Binary_Digits: in Positive; Mass,B1,B2,B3,Del: in Flt) return FRParam is
    --- Mass=5 is the simplest choice
    Fac: constant Flt := Mass/Flt(5);
    BB1: constant Flt := B1*B1;
    BB2: constant Flt := B2*B2;
    BB3: constant Flt := B3*B3;
    I1:  constant Flt := Fac*(BB2+BB3);
    I2:  constant Flt := Fac*(BB3+BB1);
    I3:  constant Flt := Fac*(BB1+BB2);
    I11,I12,I22: Flt;
    P: FRParam;
  begin
    if BB1<BB2 then raise Not_Implemented; end if; --- undesirable
    if BB2<BB3 then raise Not_Implemented; end if; --- undesirable
    Approx_Inertia(I1,I2,Del,I11,I12,I22);
    P := (I11,I12,I22,I3,BB1,BB2,BB3,Mass);
    Reduce_Precision(Binary_Digits,P);
    return P;
  end Approx_Homog_Param;

  function Approx_EMin(P: FRParam) return Flt is
    use Flt_EF;
  begin
    if P.BB1<P.BB2 then raise Not_Implemented; end if; --- undesirable
    if P.BB2<P.BB3 then raise Not_Implemented; end if; --- undesirable
    return P.Mass*GAccel*Sqrt(P.BB3);
  end Approx_EMin;

  procedure Show(P: in FRParam) is
  begin
    Show1("I11  =",P.I11);
    Show1("I12  =",P.I12);
    Show1("I22  =",P.I22);
    Show1("I33  =",P.I33);
    Show1("BB1  =",P.BB1);
    Show1("BB2  =",P.BB2);
    Show1("BB3  =",P.BB3);
    Show1("Mass =",P.Mass);
    Show1("EMin =",Approx_EMin(P));
  end Show;

  procedure Put(F: in File_Type; P: in FRParam; Decimal: in Boolean := False) is
  begin
    Put(F,P.I11,Decimal);
    Put(F,P.I12,Decimal);
    Put(F,P.I22,Decimal);
    Put(F,P.I33,Decimal);
    Put(F,P.BB1,Decimal);
    Put(F,P.BB2,Decimal);
    Put(F,P.BB3,Decimal);
    Put(F,P.Mass,Decimal);
  end Put;

  procedure Get(F: in File_Type; P: in out FRParam; Decimal: in Boolean := False) is
  begin
    Get(F,P.I11,Decimal);
    Get(F,P.I12,Decimal);
    Get(F,P.I22,Decimal);
    Get(F,P.I33,Decimal);
    Get(F,P.BB1,Decimal);
    Get(F,P.BB2,Decimal);
    Get(F,P.BB3,Decimal);
    Get(F,P.Mass,Decimal);
  end Get;

  ------------------------------------------------------------------
  ------------------------------------------------------------------

  procedure Show(N: in String; P: in Flag) is
    use Mod_IO;
  begin
    Put(Standard_Output,N);
    Put(Standard_Output,P,2);
    New_Line(Standard_Output);
  end Show;

  function Choose(ChooseF0: Boolean; F0,F1: Flag) return Flag is
  begin
    if ChooseF0 then return F0; else return F1; end if;
  end Choose;

end Params;
