with Globals, Ints, Num_Polynoms;
use Globals, Ints;

pragma Elaborate_All (Globals,Ints,Num_Polynoms);

package body Polynoms.Roots is

  use SP_Soil;

  procedure Norm(X,Y: in Scalar; S,Tmp: in out Scalar) is
  begin
    if IsZero(Y) then
      Norm(X,S);
    elsif IsZero(X) then
      Norm(Y,S);
    else
      Prod(X,X,S);
      AddProd(Y,Y,S,Tmp);
      Norm(S,Tmp);
      Sqrt(Tmp,S);
    end if;
  end Norm;

  procedure Sqrt(Zr,Zi: in Scalar; Wr,Wi: in out Scalar) is
    --- W := Sqrt(Z) where Z=Zr+i*Zi and W=Wr+i*Wi
    --- branch chosen so that Wr>=0 and sign(Wi)=sign(Zi)
    Re_Sign: constant Integer := Sign(Zr);
    Im_Sign: constant Integer := Sign(Zi);
  begin
    if Im_Sign = 0 then      --- Z is real
      if Re_Sign >= 0 then
        Sqrt(Zr,Wr);
        SetZero(Wi);
      else
        Neg(Zr,Wr);
        Sqrt(Wr,Wi);
        SetZero(Wr);
      end if;
    else
      declare
        S,Tmp: Scalar;
      begin
        Norm(Zr,Zi,Tmp,S);
        Norm(Zr,S);
        Add(S,Tmp);
        Mult(Half,Tmp,S);
        Sqrt(Tmp,S);         --- S = Sqrt(Half*(Norm(Z)+abs(Zr)));
        if Re_Sign >= 0 then
          Copy(S,Wr);
        end if;
        if Im_Sign < 0 then
          Neg(S);
        end if;
        if Re_Sign <= 0 then
          Copy(S,Wi);
        end if;
        if Re_Sign < 0 then
          Quot(Zi,Wi,Wr);
          Mult(Half,Wr,Tmp);
        end if;
        if Re_Sign > 0 then
          Quot(Zi,Wr,Wi);
          Mult(Half,Wi,Tmp);
        end if;
      end;
    end if;
  end Sqrt;

  procedure QSolve(Zr,Zi: in Scalar; Wr1,Wi1,Wr2,Wi2: in out Scalar) is
    --- W=W1 and W=W2 solve W*W-2*Z*W+1=0
    Sr: Scalar renames Wr1;
    Si: Scalar renames Wi1;
    ZZr: Scalar renames Wr2;
    ZZi: Scalar renames Wi2;
    Tmp: Scalar renames Wi2;
  begin
    Prod(Zr,Zr,ZZr);
    Prod(Zi,Zi,Tmp);
    Sub(Tmp,ZZr);
    Prod(Zr,Zi,ZZi);
    Mult(Scal_Two,ZZi);  -- ZZ = Z*Z
    Sub(Scal_One,ZZr);
    Sqrt(ZZr,ZZi,Sr,Si); -- S = Sqrt(Z*Z-1)
    Diff(Zr,Sr,Wr2);
    Diff(Zi,Si,Wi2);     -- W2 = Z-Sqrt(Z*Z-1)
    Add(Zr,Wr1);
    Add(Zi,Wi1);         -- W1 = Z+Sqrt(Z*Z-1)
    if MaxNorm(Wr2)>MaxNorm(Wr1) then
      Swap(Wr1,Wr2);
      Swap(Wi1,Wi2);
    end if;
  end QSolve;

  procedure Real_Root(P: in Polynom1; WMin,WMax: in Flt; W: in out Scalar; Eps: in Radius) is
    use Num_Polynoms;
    Deg: constant Positive := EffDeg(P);
    Err: Radius := Eps;
  begin
    if Eps=Zero then
      null;
    elsif STrunc then
      RoughRealRoot(P,WMin,WMax,W,Err);
    else
      for I in 0 .. Deg loop Put_Numeric(I+1,P(I)); end loop;
      NumRealRoot(Deg,WMin,WMax,Eps);
      Proper_Rounding;
      Get_Numeric(1,W);
    end if;
    FindRealRoot(P,W,Err);
  end Real_Root;

  procedure Real_Roots(P: in Polynom1; WMin,WMax: in Flt; W: in out Vector; Eps: in Radius) is
    use SV_Sorting, Num_Polynoms;
    Deg: constant Positive := EffDeg(P);
  begin
    if Eps=Zero then
      FindRealRoots(P,W);
    elsif STrunc then
      NumRealRoots(P,WMin,WMax,W,Eps);
      Default_Sort(W);
    else
      for I in 0 .. Deg loop Put_Numeric(I+1,P(I)); end loop;
      NumRealRoots(Deg,WMin,WMax,Eps);
      Proper_Rounding;
      for I in 1 .. Deg loop Get_Numeric(I,W(I)); end loop;
      Default_Sort(W);
      FindRealRoots(P,W);
    end if;
  end Real_Roots;

  procedure Circle_Roots(P: in Polynom1; Hr: in out Vector; Eps: in Radius) is
    --- roots of P expected on the unit circle, Hr(I) are the real parts
    Deg: constant Positive := EffDeg(P);
    Q: Polynom1(0 .. Deg/2);
  begin
    if (Deg mod 2)=1 then raise Not_Implemented; end if;
    SymCosPoly(P,Q);
    Real_Roots(Q,NegOne,One,Hr,Eps);
  end Circle_Roots;

  procedure Sym_QFactor(P: in Polynom1; UMin,UMax: in Flt; U: in out Scalar; Q: in out Polynom1; Eps: in Radius) is
    use Num_Polynoms;
    Deg: constant Positive := EffDeg(P);
    Err: Radius := Eps;
  begin
    if Eps=Zero then
      null;
    elsif STrunc then
      RoughSymQFactor(P,UMin,UMax,U,Q,Err);
    else
      for I in 0 .. Deg loop Put_Numeric(I+1,P(I)); end loop;
      NumSymQFactor(Deg,UMin,UMax,Eps);
      Proper_Rounding;
      Get_Numeric(1,U);
    end if;
    FindSymQFactor(P,U,Q,Err);
  end Sym_QFactor;

  procedure Sym_QFactor(P: in Polynom1; UMin,UMax: in Flt; U: in out Scalar; Eps: in Radius) is
    Deg: constant Natural := EffDeg(P);
    Q: Polynom1(0 .. Deg-2);
  begin
    Sym_QFactor(P,UMin,UMax,U,Q,Eps);
  end Sym_QFactor;

  procedure Symplectic_Roots(P: in Polynom1; Hr,Hi: in out Vector; Eps: in Radius) is
    --- gets rid of pair of eigenvalues near 1 before using SymCosPoly and Gen_Roots
    --- Eps is used only in Sym_QFactor
    Dt: constant Positive := EffDeg(P)-2;
    UMin: constant Flt := -Flt(1025)/Flt(512);
    UMax: constant Flt := -Flt(1023)/Flt(512);
    D: Natural := Hr'Last;
    U: Scalar;
    Pt: Polynom1(0 .. Dt);
    Q: Polynom1(0 .. Dt/2);
  begin
    if Eps=Zero then raise Not_Implemented; end if; --- need to compute U from scratch
    Sym_QFactor(P,UMin,UMax,U,Pt,Eps); --- find U near -2
    SymCosPoly(Pt,Q);
    Gen_Roots(Q,Hr,Hi,D);
    Prod(NegHalf,U,Hr(D+1));
    SetZero(Hi(D+1));
    if D<Dt/2 then
      Show1("Symplectic_Roots missing:",Dt/2-D);
      raise Sorry;
    end if;
  end Symplectic_Roots;

  procedure Gen_Roots(P: in Polynom1; Wr,Wi: in out Vector; D: out Natural) is
    --- D is the number of roots found
    use Num_Polynoms;
    Deg: constant Positive := EffDeg(P);
  begin
    D := Deg;
    for I in 0 .. Deg loop Put_Numeric(I+1,P(I)); end loop;
    NumRoots(D);
    Proper_Rounding;
    for I in 1 .. Deg loop Get_Numeric(I,Wr(I)); end loop;
    for I in 1 .. Deg loop Get_Numeric(Deg+I,Wi(I)); end loop;
    if not STrunc then
      FindRoots(P,Wr,Wi,D);
    end if;
  end Gen_Roots;

  ----------------------- for testing etc

  procedure SymProd(P1: in Polynom1; P2: in out Polynom1) is
    D: constant Natural := EffDeg(P1);
  begin
    SetZero(P2);
    for I in 0 .. D loop
      for J in 0 .. D loop
        AddProd(P1(D-I),P1(J),P2(I+J));
      end loop;
    end loop;
  end SymProd;

  function SymErr(P: Polynom1) return Radius is
    Deg: constant Natural := EffDeg(P);
    M: constant Natural := Deg/2;
    R: Radius := Zero;
    S: Scalar;
  begin
    if 2*M /= Deg then raise Undefined; end if;
    for K in reverse 1 .. M loop
      Diff(P(M+K),P(M-K),S);
      R := RMax(MaxNorm(S),R);
    end loop;
    return R;
  end SymErr;

  procedure RandomRoots(W: in out Vector; Fac: in Flt := One) is
  begin
    for I in W'Range loop Assign(Fac*Simple_Random,W(I)); end loop;
  end RandomRoots;

  procedure RandomRoots(NSimple: in Natural; Wr,Wi: in out Vector; RFac,IFac: in Flt := One) is
    Deg: constant Positive := Wr'Last;
    I: Integer := 1;
  begin
    loop
      if I <= NSimple then
        Assign(RFac*Simple_Random,Wr(I));
        SetZero(Wi(I));
        I := I+1;
      else
        Assign(RFac*Simple_Random,Wr(I));
        Copy(Wr(I),Wr(I+1));
        Assign(IFac*Simple_Random,Wi(I));
        Neg(Wi(I),Wi(I+1));
        I := I+2;
      end if;
      exit when I>Deg;
    end loop;
  end RandomRoots;

  procedure RandomReal(W: in out Vector; P: in out Polynom1; Fac: in Flt := One) is
  begin
    RandomRoots(W,Fac);
    Roots2Poly(W,P);
  end RandomReal;

  procedure RandomCC(Wr,Wi: in out Vector; P: in out Polynom1; RFac,IFac: in Flt := One) is
  begin
    Random(0,Wr,Wi,P,RFac,IFac);
  end RandomCC;

  procedure Random(NSimple: in Natural; Wr,Wi: in out Vector; P: in out Polynom1; RFac,IFac: in Flt := One) is
  begin
    RandomRoots(NSimple,Wr,Wi,RFac,IFac);
    Roots2Poly(Wr,Wi,P);
  end Random;

  procedure RandomPoly(U,V: in out Vector; P: in out Polynom1; UFac,VFac: in Flt := Three) is
  begin
    SetZero(P);
    Assign(One,P(0));
    RandomRoots(U,UFac);
    RandomRoots(V,VFac);
    for I in U'Range loop
      QMultiply(U(I),V(I),Scal_Zero,Scal_Zero,P);
    end loop;
  end RandomPoly;

  procedure RandomSymPoly(U: in out Vector; P: in out Polynom1; Fac: in Flt := Two) is
    --- Fac=2 corresponds to roots on the unit circle
  begin
    SetZero(P);
    Assign(One,P(0));
    RandomRoots(U,Fac);
    for K in U'Range loop
      QMultiply(U(K),Scal_One,Scal_Zero,Scal_Zero,P);
    end loop;
  end RandomSymPoly;


end Polynoms.Roots;
