with Globals, Strings, Ints;
use Globals, Strings, Ints;

pragma Elaborate_All (Globals,Strings,Ints);

package body Polynoms.Soil is

  ------------------------------- Newton for general polynomials

  procedure RealRoot1(P0,P1: in Scalar; W: in out Scalar) is
    use SMV;
  begin
    if IsZero(P1) then
      Show0("RealRoot1: no (isolated) root");
      raise Not_Found;
    end if;
    Quot(P0,P1,W);
    Neg(W);
  end RealRoot1;

  procedure RealRoots2(P0,P1,P2: in Scalar; W1,W2: in out Scalar) is
    use SMV;
    B,C,V: Scalar;
  begin
    Quot(P1,P2,B);
    Quot(P0,P2,C);
    Roots2(B,C,W1,W2,V);
    if not IsZero(V) then
      Show0("RealRoots2: no real root");
      raise Not_Found;
    end if;
  end RealRoots2;

  procedure FindRealRoot(P: in Poly; W: in out Scalar; Err: out Radius) is
    Deg: constant Natural := EffDeg(P);
  begin
    if Deg<3 then
      if Deg=2 then
        declare
          W1,W2,S1,S2: Scalar;
        begin
          RealRoots2(P(0),P(1),P(2),W1,W2);
          Diff(W1,W,S1);
          Diff(W2,W,S2);
          if MaxNorm(S1)<MaxNorm(S2) then
            Copy(W1,W);
          else
            Copy(W2,W);
          end if;
          Err := Zero;
        end;
      else
        RealRoot1(P(0),P(1),W);
        Err := Zero;
      end if;
    else
      declare
        Steps: constant Natural := Choose(STrunc,64,0);
        Verb: constant Natural := Verbosity;

        procedure TF(X: in Scalar; FX,DFX: in out Scalar) is
        begin
          Evaluate(P,X,FX,Deg);
          EvalDer(P,X,DFX,Deg);
          Err := MaxNorm(FX);
        end TF;

        use SMV;
        procedure Find is new Find01 (TF => TF);
      begin
        Find(W,Steps,Verb);
      exception
        when No_Contraction =>
          Show0("FindRealRoot: no contraction");
          raise Not_Found;
      end;
      if Verbosity>4 then Show1("FindRealRoot: Err =",Err); end if;
    end if;
  end FindRealRoot;

  procedure FindRealRoots(P: in Poly; W: in out Vector) is
    --- dies if there is a problem
    D: Integer := EffDeg(P);
    MaxErr: Radius := Zero;
    Err: Radius;
  begin
    if W'First /= 1 then raise Not_Implemented; end if;
    if D<3 then
      if D=2 then
        RealRoots2(P(0),P(1),P(2),W(1),W(2));
      else
        RealRoot1(P(0),P(1),W(1));
      end if;
    else
      for N in W'Range loop
        if D>0 then
          FindRealRoot(P,W(N),Err);
          if Err>MaxErr then MaxErr := Err; end if;
        end if;
        D := D-1;
      end loop;
      if D<0 and then Verbosity>1 then
        Show2("FindRealRoots warning: W'Range too large:",W'First,W'Last);
      end if;
    end if;
    if Verbosity>3 then
      Show1("FindRealRoots: MaxErr =",MaxErr);
    end if;
  end FindRealRoots;

  procedure FindQFactor2(P0,P1,P2: in Scalar; U,V: in out Scalar; Q: in out Poly) is
  begin
    if IsZero(P2) then raise Not_Found; end if;
    for I in 1 .. Q'Last loop SetZero(Q(I)); end loop;
    Copy(P2,Q(0));
    Quot(P1,P2,U);
    Quot(P0,P2,V);
  end FindQFactor2;

  procedure FindQFactor(P: in Poly; U,V: in out Scalar; Q: in out Poly; Err: out Radius) is
    --- U,V are approximate solutions
    --- if not STrunc then only error estimates are added
    DegP: constant Natural := EffDeg(P);
  begin
    if DegP<3 then
      FindQFactor2(P(0),P(1),P(2),U,V,Q);
      Err := Zero;
    else
      declare
        Steps: constant Natural := Choose(STrunc,64,0);
        Verb: constant Natural := Verbosity;
        DegQ: constant Natural := DegP-2;
        DegR: constant Integer := DegP-4;
        R: Poly(0 .. IMax(DegR,1));
        use SMV;

        procedure TF(S: in Vector2; FS: in out Vector2; DFS: in out Matrix2) is
        begin
          Bairstow0(P,DegP,S,Q,FS);
          Bairstow1(Q,DegQ,S,R,DFS);
          Err := MaxNorm(FS);
        end TF;

        S: Vector2 := (U,V);

        procedure Find is new SMV.Find02 (TF => TF);
      begin
        Find(S,Steps,Verb);   --- last step produces correct Q
        Copy(S(1),U);
        Copy(S(2),V);
      exception
        when No_Contraction =>
          Show0("PolyRoots.FindQFactor: no contraction");
          raise Not_Found;
      end;
    end if;
  end FindQFactor;

  procedure FindQFactors(P: in Poly; U,V: in out Vector; A,B: in out Scalar) is
    --- last factor is A*x+B
    --- dies if there is a problem
    Deg: constant Natural := EffDeg(P);
    MaxErr: Radius := Zero;
    Err: Radius;
    S: Scalar;
    Q: Poly(0 .. Deg);
  begin
    for N in 1 .. Deg/2 loop
      FindQFactor(P,U(N),V(N),Q,Err);
      if Err>MaxErr then MaxErr := Err; end if;
      if Verbosity>4 then
        Show1(Strng(N,4) & " FindQFactors Err =",Err);
      end if;
    end loop;
    if (Deg mod 2)=1 then
      Quot(B,A,S);
      Neg(S);
      FindRealRoot(P,S,Err);
      if Err>MaxErr then MaxErr := Err; end if;
      Assign(One,A);
      Neg(S,B);
    else
      SetZero(A);
      Assign(One,B);
    end if;
    if Verbosity>3 then
      Show1("FindRealRoots: MaxErr =",MaxErr);
    end if;
  end FindQFactors;

  ------------------------------- Newton for symplectic polynomials

  function Right_Sign(P: Poly) return Integer is
    S0: constant Integer := Sign(P(0));
    T: Scalar := P(0);
  begin
    for I in 1 .. P'Last loop Add(P(I),T); end loop; -- value at 1
    return Sign(T)*S0;
  end Right_Sign;

  function Left_Sign(P: Poly) return Integer is
    S0: constant Integer := Sign(P(0));
    Odd: Boolean := True;
    T: Scalar := P(0);
  begin
    for I in 1 .. P'Last loop
      if Odd then Sub(P(I),T); else Add(P(I),T); end if;
      Odd := not Odd;
    end loop;           -- value at -1
    return Sign(T)*S0;
  end Left_Sign;

  procedure CircleRoots(P: in Poly; Hr: in out Vector) is
    --- roots of P need to be on the unit circle
    --- dies if there is a problem
    Deg: constant Positive := EffDeg(P);
    Q: Poly(0 .. Deg/2);
  begin
    if (Deg mod 2)=1 then raise Not_Implemented; end if;
    if Left_Sign(P)<0 then raise Off_Circle; end if;
    if Right_Sign(P)<0 then raise Off_Circle; end if;
    SymCosPoly(P,Q);
    FindRealRoots(Q,Hr);
  end CircleRoots;

  procedure FindSymQFactor2(P0,P1,P2: in Scalar; U: in out Scalar; Q: in out Poly) is
    --- for degree 2
  begin
    if IsZero(P2) then
      Show0("FindSymQFactor2: no quadratic factor");
      raise Not_Found;
    end if;
    if Verbosity>2 then
      Quot(P0,P2,U);
      Sub(Scal_One,U);
      Show1("FindSymQFactor2: last |V-1| =",MaxNorm(U));
    end if;
    for I in 1 .. Q'Last loop SetZero(Q(I)); end loop;
    Copy(P2,Q(0));
    Quot(P1,P2,U);
  end FindSymQFactor2;

  procedure FindSymQFactor(P: in Poly; U0: in out Scalar; Q: in out Poly; Err: out Radius) is
    --- roots are assumed to come in pairs (r,1/r)
    --- U0 is an approximate solution (V=1)
    --- if not STrunc then only error estimates are added
    DegP: constant Natural := EffDeg(P);
  begin
    if DegP<3 then
      FindSymQFactor2(P(0),P(1),P(2),U0,Q);
      Err := Zero;
    else
      declare
        Steps: constant Natural := Choose(STrunc,64,0);
        Verb: constant Natural := Verbosity;
        DegQ: constant Natural := DegP-2;
        DegR: constant Integer := DegP-4;
        S,FS: Vector2;
        DFS: Matrix2;
        U: Scalar renames S(1);
        V: Scalar renames S(2);
        C: Scalar renames FS(1);
        D: Scalar renames FS(2);
        DC: Scalar renames DFS(1,1);
        R: Poly(0 .. IMax(DegR,1));

        procedure TF(X: in Scalar; FX,DFX: in out Scalar) is
        begin
          Copy(X,U);
          Bairstow0(P,DegP,S,Q,FS);
          Copy(C,FX);
          Bairstow1(Q,DegQ,S,R,DFS);
          Copy(DC,DFX);
        end TF;

        use SMV;
        procedure Find is new Find01 (TF => TF);
      begin
        Assign(One,V);
        Find(U0,Steps,Verb);  --- last step yields correct Q
        Err := MaxNorm(FS);
        if Verbosity>2 then
          Show2("FindSymQFactor: C D ",C,D);
        end if;
      exception
        when No_Contraction =>
          Show0("FindSymQFactor: no contraction");
          raise Not_Found;
      end;
    end if;
  end FindSymQFactor;

  procedure FindSymQFactors(P: in Poly; U: in out Vector) is
    --- dies if there is a problem
    Deg: constant Natural := EffDeg(P);
    MaxErr: Radius := Zero;
    Err: Radius;
    Q: Poly(0 .. Deg);
  begin
    if (Deg mod 2)=1 then raise Undefined; end if;  --- Deg needs to be even
    for N in 1 .. Deg/2 loop
      FindSymQFactor(P,U(N),Q,Err);
      if Err>MaxErr then MaxErr := Err; end if;
      if Verbosity>4 then
        Show1(Strng(N,4) & " Symm_FindQFactors Err =",Err);
      end if;
    end loop;
    if Verbosity>3 then
      Show1("FindSymQFactors: MaxErr =",MaxErr);
    end if;
  end FindSymQFactors;

  procedure FindRoots(P: in Poly; Wr,Wi: in out Vector; D: in Natural) is
    --- assuming cc roots are consecutive in Wr,Wi
    --- dies if there is a problem
    Deg: constant Natural := EffDeg(P);
    I: Integer := 1;
    Err: Radius;
    U,V,S: Scalar;
    Q: Poly(0 .. Deg);
  begin
    while I <= D loop
      if IsZero(Wi(I)) then
        FindRealRoot(P,Wr(I),Err);      --- Newton1
        I := I+1;
      else
        Sum(Wr(I),Wr(I+1),U);
        Diff(Wi(I),Wi(I+1),S);
        Prod(U,U,V);
        AddProd(S,S,V);
        Mult(Quarter,V);
        Neg(U);
        FindQFactor(P,U,V,Q,Err);       --- Newton2
        Roots2(U,V,Wr(I),Wr(I+1),Wi(I));
        Neg(Wi(I),Wi(I+1));
        I := I+2;
      end if;
    end loop;
  end FindRoots;

  -------------------------------  numerical
  -------------------------------  numerical
  -------------------------------  numerical

  ------------------------------- 1d bisection and 1d Newton

  procedure RoughRealRoot(P: in Poly; WMin,WMax: in Flt; W: in out Scalar; Err: in out Radius) is
    Deg: constant Natural := EffDeg(P);
    W1,W2: Scalar;
  begin
    if Deg<3 then
      if Deg=2 then
        RealRoots2(P(0),P(1),P(2),W1,W2);
        if MaxNorm(W1)>MaxNorm(W2) then
          Copy(W1,W);
        else
          Copy(W2,W);
        end if;
      else
        RealRoot1(P(0),P(1),W);
      end if;
      Err := Zero;
    else
      declare
        FW: Scalar;
        function RF(R: Flt) return Flt is
        begin
          Assign(R,W);
          Evaluate(P,W,FW,Deg);
          return Approx(FW);
        end RF;
        function SF(Wt: Scalar) return Flt is
        begin
          Evaluate(P,Wt,FW,Deg);
          return Approx(FW);
        end SF;
        procedure Rough_Root is new Bis_PSign.RoughRoot (FltF => RF, F => SF);
      begin
        Rough_Root(WMin,WMax,W,Err);
      end;
    end if;
    if Verbosity>4 then
      Show1("RoughRealRoot Err =",Err);
    end if;
  end RoughRealRoot;

  procedure NumRealRoot(P: in Poly; WMin,WMax: in Flt; W: in out Scalar; Eps: in Radius) is
    Err: Radius := Eps;
  begin
    RoughRealRoot(P,WMin,WMax,W,Err);
    FindRealRoot(P,W,Err);
  end NumRealRoot;

  procedure NumRealRoots(P: in Poly; WMin,WMax: in Flt; W: in out Vector; Eps: in Radius) is
    --- dies if there is a problem
    Deg: constant Natural := EffDeg(P);
    Q: Poly(0 .. Deg);
  begin
    Copy(P,Q);
    if Deg<2 then --- ignore WMin,WMax
      if Deg=0 then raise Undefined; end if;
      Quot(P(0),P(1),W(1));
      Neg(W(1));
      Swap(Q(1),Q(0));
      SetZero(Q(1));
    else
      declare
        Err: Radius;
        R: Scalar;
      begin
        for N in 1 .. Deg loop
          Err := Eps;
          RoughRealRoot(Q,WMin,WMax,W(N),Err);
          if Err>Eps then
            if Verbosity>1 then Show1("NumRealRoots: Rough Err =",Err); end if;
            SetZero(W(N));
            return;
          end if;
          FindRealRoot(Q,W(N),Err);
          if Err>Eps then
            if Verbosity>1 then
              Show1("NumRealRoots: Find Err =",Err);
            end if;
            SetZero(W(N));
            return;
          end if;
          LDivision(Q,W(N),R);
        end loop;
      end;
    end if;
    FindRealRoots(P,W);
  end NumRealRoots;

  ------------------------------- symplectic roots on the circle

  procedure NumCircleRoots(P: in Poly; Hr: in out Vector; Eps: in Radius) is
    --- roots of P need to be on the unit circle
    Deg: constant Positive := EffDeg(P);
    Q: Poly(0 .. Deg/2);
  begin
    if (Deg mod 2)=1 then raise Not_Implemented; end if;
    SymCosPoly(P,Q);
    NumRealRoots(Q,NegOne,One,Hr,Eps);
  end NumCircleRoots;

  procedure RoughSymQFactor(P: in Poly; UMin,UMax: in Flt; U0: in out Scalar; Tmp: in out Poly; Err: in out Radius) is
    Deg: constant Natural := EffDeg(P);
    Q: Poly renames Tmp;
  begin
    if Deg<3 then
      FindSymQFactor2(P(0),P(1),P(2),U0,Q);
      Err := Zero;
    else
      declare
        S,FS: Vector2;
        U: Scalar renames S(1);
        C: Scalar renames FS(1);
        function RF(Ut: Flt) return Flt is
        begin
          Assign(Ut,U);
          Bairstow0(P,Deg,S,Q,FS);
          return Approx(C);
        end RF;
        function SF(Ut: Scalar) return Flt is
        begin
          Copy(Ut,U);
          Bairstow0(P,Deg,S,Q,FS);
          return Approx(C);
        end SF;
        procedure Rough_Root is new Bis_PSign.RoughRoot (FltF => RF, F => SF);
      begin
        Assign(One,S(2));
        Rough_Root(UMin,UMax,U,Err);
        Copy(U,U0);
      end;
    end if;
    if Verbosity>4 then
      Show1("RoughSymQFactor Err =",Err);
    end if;
  end RoughSymQFactor;

  procedure NumSymQFactor(P: in Poly; UMin,UMax: in Flt; U: in out Scalar; Eps: in Radius) is
    Deg: constant Natural := EffDeg(P);
    Err: Radius := Eps;
    Q: Poly(0 .. Deg-2);
  begin
    RoughSymQFactor(P,UMin,UMax,U,Q,Err);
    FindSymQFactor(P,U,Q,Err);
  end NumSymQFactor;

  procedure NumSymQFactors(P: in Poly; UMin,UMax: in Flt; U: in out Vector; Eps: in Radius) is
    Deg: constant Natural := EffDeg(P);
    Err: Radius := Eps;
    Pt,Q: Poly(0 .. Deg);
  begin
    Copy(P,Q);
    if (Deg mod 2)=1 then raise Undefined; end if;
    for N in 1 .. Deg/2 loop
      Copy(Q,Pt);
      Err := Eps;
      RoughSymQFactor(Pt,UMin,UMax,U(N),Q,Err);
      if Err>Eps then
        if Verbosity>1 then Show1("NumSymQFactors: Rough Err =",Err); end if;
        SetZero(U(N));
        return;
      end if;
      FindSymQFactor(Pt,U(N),Q,Err);
      if Err>Eps then
        if Verbosity>1 then Show1("NumSymQFactors: Find Err =",Err); end if;
        SetZero(U(N));
        return;
      end if;
    end loop;
    FindSymQFactors(P,U);
  end NumSymQFactors;





end Polynoms.Soil;
