with Ints, Ada.Strings.Unbounded, Pooling, ObO;
use Ints;

pragma Elaborate_All (Ints,Ada.Strings.Unbounded,Pooling,ObO);

generic

  PDeg: in Positive;           -- Taylor degree
  Prec: in Positive;           -- desired Precision for Integrate
  NegLogMinRad: in Positive;   -- MinRad := Half**NegLogMinRad
  MaxRadFac: in Positive := 8; -- MaxRad := MaxRadFac*MinRad

package Rattleback.Flows is

  MinRad: constant Radius := Half**NegLogMinRad;    -- smallest time increment
  MaxRadFacPlus1: constant Positive := MaxRadFac+1;

  package RObO is new ObO (PDeg => PDeg, Scalar => Scalar);
  use RObO;

  procedure Show_Rattle_Param;

  subtype TVector3 is TVector(OneTwoThree);
  subtype TMatrix3 is TMatrix(OneTwoThree,OneTwoThree);

  type Curve is
    record
      Alpha,Beta,Gamma,M: TVector3;
    end record;
  procedure Copy(C1: in Curve; C2: in out Curve);

  procedure Val(R: in Radius; U: in TVector3; S: in Scalar; V: in out SVector3);
  procedure Val(R: in Radius; C: in Curve; S: in Scalar; P: in out Point);

  ---------- order by order
  procedure Diag_Prod(Deg: in Power; B: in SVector3; U: in TVector3; V: in out TVector3);
  procedure TV_Prod(R: in Radius; Deg: in Power; P: in Taylor; U: in TVector3; V: in out TVector3);
  procedure VV_Prod(R: in Radius; Deg: in Power; U,V: in TVector3; W: in out TVector3);
  subtype Data_Dot is TVector(1..3);
  procedure Dot(R: in Radius; Deg: in Power; O: in out Data_Dot; U,V: in TVector3; P: in out Taylor);
  subtype Data_Cross is TVector(1..6);
  procedure Cross1(R: in Radius; Deg: in Power; O: in out Data_Cross; U,V: in TVector3; W: in out TVector3);
  procedure Cross2(R: in Radius; Deg: in Power; O1,O2: in out Data_Cross; U1,U2,V1,V2: in TVector3; W1,W2: in out TVector3);
  procedure Cross3(R: in Radius; Deg: in Power; O1,O2,O3: in out Data_Cross; U1,U2,U3,V1,V2,V3: in TVector3; W1,W2,W3: in out TVector3);
  procedure SM_AddProd(Deg: in Power; A: in SMatrix3; S: in Scalar; B: in TMatrix3; C: in out TMatrix3);
  procedure Symm_TM_Prod(R: in Radius; Deg: in Power; P: in Taylor; A: in TMatrix3; B: in out TMatrix3);
  subtype Data_MV_Prod is TVector(1..9);
  procedure MV_Prod(R: in Radius; Deg: in Power; O: in out Data_MV_Prod; A: in TMatrix3; U: in TVector3; V: in out TVector3);
  subtype Data_Sandwich is TVector(1..6);
  procedure Sandwich(R: in Radius; Deg: in Power; O: in out Data_Sandwich; U: in TVector3; A: in out TMatrix3);
  subtype Data_Det is TVector(1..3);
  procedure Symm_Det(R: in Radius; Deg: in Power; O: in out Data_Det; A,B: in TMatrix3; P: in out Taylor);
  subtype Data_Adj is TVector(1..12);
  procedure Symm_Adj(R: in Radius; Deg: in Power; O: in out Data_Adj; A: in TMatrix3; B: in out TMatrix3);
  type Data_M_Inv is private;
  procedure Copy(O1: in Data_M_Inv; O2: in out Data_M_Inv);
  procedure Symm_M_Inv(R: in Radius; Deg: in Power; O: in out Data_M_Inv; A: in TMatrix3; B: in out TMatrix3);

  type Data_VecField is private;
  procedure Copy(O1: in Data_VecField; O2: in out Data_VecField);
  procedure VecField(R: in Radius; Deg: in Power; O: in out Data_VecField; X: in Curve; DotX: in out Curve);

  procedure Low_AntiDer(Deg: in Power; DotU: in TVector3; U0: in SVector3; U: in out TVector3);
  procedure Low_AntiDer(Deg: in Power; DotX: in Curve; X0: in Point; X: in out Curve);
  procedure Hi_AntiDer(R: in Radius; DotU: in TVector3; U: in out TVector3; Tmp: in out SVector3; Contr: in out Boolean);
  procedure Hi_AntiDer(R: in Radius; DotX: in Curve; X: in out Curve; Tmp: in out SVector3; Contr: in out Boolean);
  procedure Hi_ErrMult(Fac: in Radius; U: in out TVector3);
  procedure Hi_ErrMult(Fac: in Radius; X: in out Curve);

  ---------- integration
  type RData is private;
  procedure Show_Stat(D: in RData);
  procedure Update(C: in Curve; D: in out RData);

  procedure Integrate(X0: in Point; X,DotX: in out Curve; VTmp: in out SVector3; D: in out RData);
  procedure Prepare_Plot(T1,T2: in Flt; FileName: in String);
  procedure Show_Info(T: in Flt; P: in Point);

  YawSteps:  Positive := 2**5; --- best a power of 2
  RollSteps: Positive := 2**5; --- best a power of 2

  procedure YawNumPi(R: in Radius; T: in Scalar; X,DotX: in Curve; NY: in out Integer);  --- uses YawSteps
  procedure RollNumPi(R: in Radius; T: in Scalar; X,DotX: in Curve; NR: in out Integer); --- uses RollSteps
  procedure NumPi(R: in Radius; T: in Scalar; X,DotX: in Curve; NY,NR: in out Integer);

  generic
    with function Done(Tau,DTau: in Flt; C: in Curve) return Boolean;
  procedure Generic_Flow(T,DT: out Flt; XT: in out Point; X,DotX: in out Curve; Forward: in Boolean := True);

  procedure Plain_Flow(TNeed: in Flt; T,DT: out Flt; XT: in out Point; X,DotX: in out Curve);
  procedure TimeT_Map(TStop: in Scalar; X0: in Point; XT: in out Point);

  ---------- Poincare
  procedure Val(F: in Flt_Affine; V: in TVector3; S: in out Taylor);
  procedure Plain_Poincare(T1,T2: in Flt; F: in Flt_Affine; X0: in Point; XT: in out Point);
  function Sign_Change(DT: in Flt; F: in Flt_Affine; M: in TVector3; Tmp: in out Taylor) return Boolean;
  procedure Sign_Poincare(TNeed: in Flt; F: in Flt_Affine; X0: in Point; XT: in out Point);
  function Neg2Pos(DT: in Flt; F: in Flt_Affine; M: in TVector3; Tmp: in out Taylor) return Boolean;
  procedure Neg2Pos_Poincare(TNeed: in Flt; F: in Flt_Affine; X0: in Point; XT: in out Point);
  procedure Pos2Neg_Poincare(TNeed: in Flt; F: in Flt_Affine; X0: in Point; XT: in out Point);
  procedure CountNeg2Pos_Poincare(Iter: in Integer; F: in Flt_Affine; X0: in Point; XT: in out Point);
  procedure CountPos2Neg_Poincare(Iter: in Integer; F: in Flt_Affine; X0: in Point; XT: in out Point);

  Allow_Parallel: Boolean := True;

  --- should be private
  procedure Init_Plot;
  procedure Plot(T: in Flt; P: in Point);
  procedure Stop_Plot;

private

  Rigorous_Alpha:         constant Boolean := Compute_Alpha and Not_STrunc;
  Rigorous_Beta:          constant Boolean := Compute_Beta and Not_STrunc;

  type Data_M_Inv is
    record
      Det,InvDet: Taylor;
      Adj: TMatrix3;
      Det1: Data_Det;
      Inv1: Data_Inv;
      Adj1: Data_Adj;
    end record;

  type Data_VecField is
    record
      SS,InvSS,InvS,SDotS,InvSDotS: Taylor;
      BBGamma,BBDotGamma,SDotR,Omega,VTmp1,VTmp2,VTmp3,VTmp4,DotM1,DotM2,DotM3: TVector3;
      SSK,K,ImK,InvImK: TMatrix3;
      Inv1: Data_Inv;
      Sqrt1: Data_Sqrt;
      Dot1,Dot2: Data_Dot;
      Cross1,Cross2,Cross3,Cross4,Cross5,Cross6,Cross7: Data_Cross;
      Inv2: Data_M_Inv;
      Prod1: Data_MV_Prod;
      Sand1: Data_Sandwich;
    end record;

  subtype Counter is Int_Vec(0 .. MaxRadFacPlus1);
  type RData is
    record
      Rad: Radius;
      Count: Counter := (others => 0);
    end record;

  function MaxCoeffPDegMinus1(C: Curve) return Radius;
  procedure ShowTime(T0: in Flt; DT: in Scalar; Tmp: in out Scalar);

  ---------- plotting
  package ASU renames Ada.Strings.Unbounded;
  Plot_TMin: Flt := Zero;
  Plot_TMax: Flt := Zero;
  Plot_Filename: ASU.Unbounded_String := ASU.Null_Unbounded_String;
  Plot_Filetype: File_Type;

  ---------- pooling
  procedure CP(O1: in Data_VecField; O2: in out Data_VecField) renames Copy;
  package XP is new Pooling (Item => Data_VecField);
  subtype Pooled_Data_VecField is XP.Controlled_LD_Pointer; --- not used here

  package XAlloc renames XP.Alloc;
  subtype XT_Pointer is XAlloc.LD_Pointer;
  XAC: XAlloc.Controlled_Protected_Data;
  XPool: XAlloc.Protected_Data renames XAC.P;

end Rattleback.Flows;
