with Ada.Text_IO, Strings, Ints, Flts, Globals, Params;
use Ada.Text_IO, Strings, Ints, Flts, Globals, Params;
with Flts.Std.Balls, MPFR.Floats.Balls, Rattleback, Newton;

pragma Elaborate_All (Ada.Text_IO,Strings,Ints,Flts,Globals,Params);
pragma Elaborate_All (Flts.Std.Balls,MPFR.Floats.Balls,Rattleback,Newton);

procedure RSp_Stat is --- check RS'-symmetric stationary solutions

  FRP:     constant FRParam  := Param13; -- rattleback parameters
  NDigits: constant Positive := 256;     -- number of binary digits, for Find_Zeros
  Dummy:   constant Positive := MPFR.Floats.Set_Precision(NDigits);

  subtype Ball is MPFR.Floats.Balls.Ball;

  procedure Show_Gamma_Quot(Theta: in Ball) is
    use MPFR.Floats.Balls;
    ThetaHalf,B1,B2,S,C,Q: Ball;
  begin
    Assign(FRP.BB1,Q);
    Sqrt(Q,B1);
    Assign(FRP.BB2,Q);
    Sqrt(Q,B2);
    Prod(Half,Theta,ThetaHalf);
    Sin(ThetaHalf,S);
    Cos(ThetaHalf,C);
    Mult(B2,S);
    Mult(B1,C);
    Quot(S,C,Q);
    Show1("G1/G2 =",Q);
  end Show_Gamma_Quot;

  procedure Find_Zeros is
    use MPFR.Floats.Balls;
    package RB is new Rattleback(FRP => FRP, Scalar => Ball);
    procedure TQ(X: in Ball; Y: in out Ball; L: in out Ball) is begin RB.TQFun(X,Y,L); end TQ;
    package Newt is new Newton(Numeric => False, Scalar => Ball, Operator => Ball);
    procedure QFind0 is new Newt.Find0(TF => TQ);
    Steps:  constant Positive := 64;
    Theta_Star:  constant Flt := -0.2429510014810140; -- approx
    Theta_Prime: constant Flt :=  3.1305677260941773; -- approx
    Theta: Ball;
  begin
    Assign(Theta_Star,Theta);
    QFind0(Theta,Steps,Verbosity); -- Theta_Star via Newton
    Show1("Theta_Star:",Theta);
    Show_Gamma_Quot(Theta);
    Assign(Theta_Prime,Theta);
    QFind0(Theta,Steps,Verbosity); -- Theta_Prime via Newton
    Show1("Theta_Prime:",Theta);
    Show_Gamma_Quot(Theta);
  end Find_Zeros;

  subtype Scalar is Flts.Std.Balls.Ball;
  use Flts.Std.Balls;             -- no need for high precision
  package RB is new Rattleback(FRP => FRP, Scalar => Scalar);

  procedure Check_QSign(ThetaMin,ThetaMax: in Flt; Sgn: in Integer) is -- brute force
    Steps: constant Positive := 2**17;
    M: constant Positive := Steps/16;
    Theta1,Theta2: Flt;
    Theta,Q: Scalar;
  begin
    Show2("Check_QSign on ",ThetaMin,ThetaMax);
    Show1("number of subintervals",Steps);
    Theta2 := ThetaMin;
    for I in 1 .. Steps loop
      if Verbosity>0 and then (I mod M)=0 then
        Show1("step",I);
      end if;
      Theta1 := Theta2;
      Theta2 := Between(ThetaMin,ThetaMax,I,Steps);
      Enclose(Theta1,Theta2,Theta);
      RB.QFun(Theta,Q);
      if Sign(Q)/=Sgn then
        Show1("Q =",Q);
        raise Sorry;
      end if;
    end loop;
    Show1("sign always",Sgn);
  end Check_QSign;

  procedure Check_DQSign(ThetaMin,ThetaMax: in Flt; Sgn: in Integer) is -- brute force
    Steps: constant Positive := 2**17;
    M: constant Positive := Steps/16;
    Theta1,Theta2: Flt;
    Theta,Q,DQ: Scalar;
  begin
    Show2("Check_DQSign on ",ThetaMin,ThetaMax);
    Show1("number of subintervals",Steps);
    Theta2 := ThetaMin;
    for I in 1 .. Steps loop
      if Verbosity>0 and then (I mod M)=0 then
        Show1("step",I);
      end if;
      Theta1 := Theta2;
      Theta2 := Between(ThetaMin,ThetaMax,I,Steps);
      Enclose(Theta1,Theta2,Theta);
      RB.TQFun(Theta,Q,DQ);
      if Sign(DQ)/=Sgn then
        Show1("DQ =",DQ);
        raise Sorry;
      end if;
    end loop;
    Show1("sign always",Sgn);
  end Check_DQSign;

  procedure Check_Signs is
    PiUp: constant Flt := Flt(3294199)/Flt(1048576); --- upper bound on Pi
  begin
    Check_QSign(-PiUp,-One,1);
    Check_QSign(One,Two,-1);
    Check_DQSign(-One,One,-1);
    Check_DQSign(Two,PiUp,1);
  end Check_Signs;

begin
  Verbosity := 1;

  Check_Signs;
  Find_Zeros;

end RSp_Stat;
