with Ada.Text_IO, Strings, Ints, Flts, Globals, Params, Reps.Balls;
use Ada.Text_IO, Strings, Ints, Flts, Globals, Params, Reps.Balls;
with Tangents, Rattleback, Rattleback.Flows;

pragma Elaborate_All (Ada.Text_IO,Strings,Ints,Flts,Globals,Params,Reps.Balls);
pragma Elaborate_All (Tangents,Rattleback.Flows);

procedure RSpR_Der is --- derivative of G1_initial --> G3_final for P13-RSpR

  FRP:     constant FRParam  := Param13;   -- rattleback parameters
  ABFlag:  constant Flag     := 0;         -- no need to compute Alpha or Beta
  Rho:     constant Radius   := One;       -- for Tangents
  NDigits: constant Positive := 3072;      -- number of binary digits
  Dummy:   constant Positive := Reps.Set_Precision(NDigits);

  subtype Scalar is Ball;
  package Tangs is new Tangents (Dim => 1, Rho => Rho, Scalar => Scalar);
  package SRB is new Rattleback(FRP => FRP, ABFlag => ABFlag, Scalar => Scalar);
  use Tangs;
  package TRB is new Rattleback(FRP => FRP, ABFlag => ABFlag, Scalar => Tangent);

  PDeg:         constant Positive := 767;  -- Taylor degree
  Prec:         constant Positive := 1500; -- target precision for Integrate
  NegLogMinRad: constant Positive := 9;    -- MinRad := Half**NegLogMinRad
  MaxRadFac:    constant Positive := 64;   -- MaxRad := MaxRadFac*MinRad

  use TRB;
  package TRBFlows is new TRB.Flows(PDeg => PDeg, Prec => Prec, NegLogMinRad => NegLogMinRad, MaxRadFac => MaxRadFac);
  use TRBFlows;

  FileName: constant String     := "P13-RSpR.0221";      -- determines starting point
  F:        constant Flt_Affine := (Zero,Zero,Zero,One); -- Poincare section at M3=0
  TNeed:    constant Flt        := 63.0;                 -- Poincare section tau'
  NT:       constant Positive   := 7;                    -- max number of concurrent tasks

  LogFac:   constant Positive   := 256; -- Half**LogFac is the size of the initial tangent vectors

  procedure Adjust_G1(G1: in out Scalar) is
    R_DG1: constant Flt := TwoPower(-2500); -- approx 8E-904
    DG1: Scalar;
  begin
    Show0("adding [-DG1,DG1] to G1");
    DiskAt0(R_DG1,DG1);
    Add(DG1,G1);
  end Adjust_G1;

  procedure Make_Initial(P: in out Point) is
    --- RS'-symmetry has G3=M3=0
    R_G2: constant Flt := -125174.0*Half**17;
    TSize: constant Radius := TwoPower(-LogFac);
    G1,G2,E: Tangent;
    SP: SRB.Point;
  begin
    SRB.Read(Input_Dir & FileName,SP);
    Copy(SP.Gamma(1),G1(0));
    Adjust_G1(G1(0));
    Assign(TSize,G1(1));
    Assign(R_G2,G2);
    MakeABG3(G1,G2,P.Alpha,P.Beta,P.Gamma);
    Assign(Zero,Zero,Zero,P.M);
    Ham(P,E);                --- E is determined by (G,M), just for info
    Copy(E,P.Energy);
  end Make_Initial;

  procedure Save_Der(P: in Point) is
    InvTSize: constant Radius := TwoPower(LogFac);
    SDer: Scalar;
  begin
    Prod(InvTSize,P.Gamma(3)(1),SDer);
    Write(Output_Dir & FileName & ".Der",SDer);
    Reduce_Precision(64,SDer);
    Show1("Der =",SDer);
  end Save_Der;

  P0,P1: Point;

begin
  NTasks := NT;
  Verbosity := 3;

  Show0("running RSpR_Der with");
  Show_Rattle_Param;
  Make_Initial(P0);
  ShowApprox1("P0",P0);
  Pos2Neg_Poincare(TNeed,F,P0,P1);
  ShowApprox1("P1",P1);
  Save_Der(P1);

end RSpR_Der;
