with Globals, Quick_Sort;
use Globals;

pragma Elaborate_All (Globals,Quick_Sort);

package body ScalVectors.Eigenv is

  use SP;

  procedure CheckPolyDim(D: in Natural; P: in Polynom1) is
  begin
    if not (P'First=0 and P'Last=D) then
      raise Dimension_Error with "ScalVectors.Eigenv.CheckPolyDim error";
    end if;
  end CheckPolyDim;

  procedure CharPolyTri(A: in Matrix; P: in out Polynom1) is
    M: constant Natural := A'Last(1);
    S: Scalar;
    P1,P2: Polynom1(0 .. M);
  begin
    if CheckRange then
      CheckDim(M,A);
      CheckPolyDim(M,P);
    end if;
    SetZero(P1);
    SetZero(P2);
    Copy(A(1,1),P1(0));
    Assign(-1,P1(1));
    Assign(1,P2(0));
    for J in 2 .. M loop
      Prod(A(J,J-1),A(J-1,J),S);
      Mult(S,Vector(P2));
      Prod1(A(J,J),P1,P);
      Sub(P2,P);          -- P := (A(J,J)-x)*P1-A(J,J-1)*A(J-1,J)*P2
      Copy(P1,P2);
      Copy(P,P1);
    end loop;
  end CharPolyTri;

  procedure EigenPolyHess(A: in Matrix; E: in out EigenPolys) is
    --- A is assumed to be lower Hessenberg: A(I,J)=0 for J-I>1
    --- expecting dimensions A(1..N,1..N) and E(0..N,0..N)
    N: constant Positive := A'Last(1);
    S: Scalar;
  begin
    SetZero(E);
    Assign(1,E(0,0));
    for I in 1 .. N loop
      for J in 1 .. I loop
        Neg(A(I,J),S);
        for K in 0 .. J-1 loop
          AddProd(S,E(J-1,K),E(I,K));
        end loop;
      end loop;
      for K in 1 .. I loop
        Add(E(I-1,K-1),E(I,K));
      end loop;
      if I<N then
        Inv(A(I,I+1),S);
        for K in 0 .. I loop
          Mult(S,E(I,K));
        end loop;
      end if;
    end loop;
  end EigenPolyHess;

  procedure GetCharPoly(E: in EigenPolys; P: in out Polynom1) is
    --- expecting E(0..N,0..N) and P(0..N)
  begin
    GetRow(E'Last(1),E,Vector(P));
  end GetCharPoly;

  procedure QSolve(Hr,Hi: in Vector; Wr,Wi: in out Vector) is
    --- solve (W+1/W)/2=H
    use SP_Roots;
    Dh: constant Positive := Hr'Last;
    Dw: constant Positive := Wr'Last;
    Tmp1,Tmp2: Scalar;
  begin
    if Dw=Dh then
      for I in Hr'Range loop
        QSolve(Hr(I),Hi(I),Wr(I),Wi(I),Tmp1,Tmp2);
      end loop;
    elsif Dw=2*Dh then
      for I in Hr'Range loop
        QSolve(Hr(I),Hi(I),Wr(2*I-1),Wi(2*I-1),Wr(2*I),Wi(2*I));
      end loop;
    else
      raise Not_Implemented;
    end if;
  end QSolve;

  ------------------------ these need approx eigenvalues

  procedure EigenSymm(A: in Matrix; W: in out Vector) is
  begin
    EigenSymm(A,Zero,Zero,W,Zero); --- needs approx W
  end EigenSymm;

  procedure EigenSymm(A: in Matrix; W: in out Vector; V: in out Matrix) is
  begin
    EigenSymm(A,Zero,Zero,W,V,Zero); --- needs approx W
  end EigenSymm;

  procedure EigenCircle(A: in Matrix; Hr: in out Vector) is
  begin
    EigenCircle(A,Hr,Zero);  --- needs approx Hr
  end EigenCircle;

  ------------------------ these use SP_Roots

  procedure EigenSymm(A: in Matrix; WMin,WMax: in Flt; W: in out Vector; Eps: in Radius) is
    Pow4: constant Positive := 3;
    N: constant Positive := A'Last(1);
    W_Min: Flt := WMin;
    W_Max: Flt := WMax;
    S: Scalar;
    P: Polynom1(0..N);
    B,Q: Matrix(1..N,1..N);
    use SP_Roots;
  begin
    if Eps>Zero and then WMin=WMax then
      Copy(A,B);
      MaxEigen(B,Q,S,Pow4);
      W_Max := Sup(S);
      W_Min := -W_Max;
    end if;
    Copy(A,B);
    Francis(B,Q,True);
    CharPolyTri(B,P);
    Real_Roots(P,W_Min,W_Max,W,Eps);
  end EigenSymm;

  procedure EigenSymm(A: in Matrix; W: in out Vector; Eps: in Radius) is
  begin
    EigenSymm(A,Zero,Zero,W,Eps);
  end EigenSymm;

  procedure EigenSymm(A: in Matrix; WMin,WMax: in Flt; W: in out Vector; V: in out Matrix; Eps: in Radius) is
    Pow4: constant Positive := 3;
    N: constant Positive := A'Last(1);
    W_Min: Flt := WMin;
    W_Max: Flt := WMax;
    S: Scalar;
    P: Polynom1(0..N);
    B,Q: Matrix(1..N,1..N);
    E: EigenPolys(0..N,0..N);
    use SP_Roots;
  begin
    if Eps>Zero and then WMin=WMax then
      Copy(A,B);
      MaxEigen(B,Q,S,Pow4);
      W_Max := Sup(S);
      W_Min := -W_Max;
    end if;
    Copy(A,B);
    Francis(B,Q,True);
    EigenPolyHess(B,E);
    GetRow(N,E,Vector(P));
    Real_Roots(P,W_Min,W_Max,W,Eps);
    for I in 1 .. N loop
      GetRow(I-1,E,Vector(P));
      for J in 1 .. N loop
        Evaluate(P,W(J),V(I,J));
      end loop;
    end loop;
    Copy(V,B);
    Prod(Q,B,V);
  end EigenSymm;

  procedure EigenSymm(A: in Matrix; W: in out Vector; V: in out Matrix; Eps: in Radius) is
  begin
    EigenSymm(A,Zero,Zero,W,V,Eps);
  end EigenSymm;

  procedure EigenCircle(A: in Matrix; Hr: in out Vector; Eps: in Radius) is
    N: constant Positive := A'Last(1);
    P: Polynom1(0..N);
    B,Q: Matrix(1..N,1..N);
    E: EigenPolys(0..N,0..N);
    use SP_Roots;
  begin
    Copy(A,B);
    Francis(B,Q,False);
    EigenPolyHess(B,E);
    GetRow(N,E,Vector(P));
    Circle_Roots(P,Hr,Eps);
  end EigenCircle;

  procedure EigenSymplectic(A: in Matrix; Wr,Wi: in out Vector; Vr,Vi: in out Matrix; Eps: in Radius) is
    --- A(1..N,1..N), Wr(1..M), Wi(1..M), Vr(1..N,1..M), Vi(1..N,1..M)
    --- where N=M or N=2*M
    use SP_Roots;
    M: constant Positive := Wr'Last;
    N: constant Positive := A'Last(1);
    Nh: constant Positive := N/2;
    Hr,Hi: Vector(1..Nh);
    P: Polynom1(0..N);
    Tmp: Matrix(1..N,1..M);
    B,Q: Matrix(1..N,1..N);
    E: EigenPolys(0..N,0..N);
  begin
    Copy(A,B);
    Francis(B,Q,False);
    EigenPolyHess(B,E);
    GetRow(N,E,Vector(P));
    Symplectic_Roots(P,Hr,Hi,Eps);
    QSolve(Hr,Hi,Wr,Wi);
    for I in 1 .. N loop
      GetRow(I-1,E,Vector(P));
      for J in 1 .. M loop
        Evaluate(P,Wr(J),Wi(J),Vr(I,J),Vi(I,J));
      end loop;
    end loop;
    Mult(Q,Vr,Tmp);
    Mult(Q,Vi,Tmp);
  end EigenSymplectic;

  ---------------------------- numerical

  procedure SortEigen(W: in out Vector; V: in out Matrix) is
    use SV_Sorting;
    K: Integer;
    procedure Swap(I,J: in out Integer) is begin K := I; I := J; J := K; end Swap;
    function GT(I,J: Integer) return Boolean is begin return GT_MaxNorm(W(I),W(J)); end GT;
    procedure QSort is new Quick_Sort (Item => Integer, Item_Vector => Int_Vec, "<" => GT);
    P: Pivot(W'Range);
    Tmp: Vector(W'Range);
  begin
    for I in P'Range loop P(I) := I; end loop;
    QSort(P'First,P'Last,P);
    Permute(P,W,Tmp);
    Column_Permute(P,V,Tmp);
  end SortEigen;

  procedure SortEigen(Eps: in Flt; Wr,Wi: in out Vector; Vr,Vi: in out Matrix) is
    use SV_Sorting;
    K: Integer;
    procedure Swap(I,J: in out Integer) is begin K := I; I := J; J := K; end Swap;
    function GT(I,J: Integer) return Boolean is begin return GT_MaxNorm(Eps,Wr,Wi,I,J); end GT;
    procedure QSort is new Quick_Sort (Item => Integer, Item_Vector => Int_Vec, "<" => GT);
    P: Pivot(Wr'Range);
    Tmp: Vector(Wr'Range);
  begin
    for I in P'Range loop P(I) := I; end loop;
    QSort(P'First,P'Last,P);
    Permute(P,Wr,Tmp);
    Permute(P,Wi,Tmp);
    Column_Permute(P,Vr,Tmp);
    Column_Permute(P,Vi,Tmp);
  end SortEigen;

  procedure EigenQuality(A: in Matrix; W: in Scalar; V: in Vector) is
  begin
    Show1(" err ",MaxNorm(Norm1(A*V-W*V)));
  end EigenQuality;

  procedure EigenQuality(A: in Matrix; Wr,Wi: in Scalar; Vr,Vi: in Vector) is
    Nr: constant Radius := MaxNorm(Norm1(A*Vr-(Wr*Vr-Wi*Vi)));
    Ni: constant Radius := MaxNorm(Norm1(A*Vi-(Wr*Vi+Wi*Vr)));
  begin
    Show2(" err: Re Im ",Nr,Ni);
  end EigenQuality;

  procedure EigenQuality(Ar,Ai: in Matrix; Wr,Wi: in Scalar; Vr,Vi: in Vector) is
    Nr: constant Radius := MaxNorm(Norm1(Ar*Vr-Ai*Vi-(Wr*Vr-Wi*Vi)));
    Ni: constant Radius := MaxNorm(Norm1(Ar*Vi+Ai*Vr-(Wr*Vi+Wi*Vr)));
  begin
    Show2(" err: Re Im ",Nr,Ni);
  end EigenQuality;

  procedure EigenQuality(A: in Matrix; W: in Vector; V: in Matrix) is
  begin
    for K in W'Range loop
      Show1("eigen",K,False);
      Show1(" ",Approx(W(K)),False);
      EigenQuality(A,W(K),GetColumn(K,V));
    end loop;
  end EigenQuality;

  procedure EigenQuality(A: in Matrix; Wr,Wi: in Vector; Vr,Vi: in Matrix) is
  begin
    for K in Wr'Range loop
      Show1("eigen",K,False);
      EigenQuality(A,Wr(K),Wi(K),GetColumn(K,Vr),GetColumn(K,Vi));
    end loop;
  end EigenQuality;

  procedure EigenQuality(Ar,Ai: in Matrix; Wr,Wi: in Vector; Vr,Vi: in Matrix) is
  begin
    for K in Wr'Range loop
      Show1("eigen",K,False);
      EigenQuality(Ar,Ai,Wr(K),Wi(K),GetColumn(K,Vr),GetColumn(K,Vi));
    end loop;
  end EigenQuality;

end ScalVectors.Eigenv;
